#!/usr/bin/env python
"""
__init__.py. Includes wrapper for running dyPolyChord with different
dynamic goals.
"""
from dyPolyChord._version import __version__
from dyPolyChord.run_dynamic_ns import run_dypolychord



#!/usr/bin/env python
"""
Functions for running dyPolyChord using compiled PolyChord C++ or Fortran
likelihoods.
"""
import os


class RunCompiledPolyChord(object):

    """Object for running a compiled PolyChord executable with specified
    inputs."""

    def __init__(self, executable_path, prior_str, **kwargs):
        """
        Specify path to executable, priors and derived parameters.

        Parameters
        ----------
        executable_path: str
            Path to compiled likelihood. If this is in the directory from which
            dyPolyChord is being run, you may need to prepend "./" to the
            executable name for it to work.
        prior_str: str
            String specifying prior in the format required for PolyChord .ini
            files (see get_prior_block_str for more details).
        config_str: str, optional
            String to be written to [root].cfg file if required.
        derived_str: str or None, optional
            String specifying prior in the format required for PolyChord .ini
            files (see prior_str for more details).
        mpi_str: str or None, optional
            Optional mpi command to preprend to run command.
            For example to run with 8 processors, use mpi_str = 'mprun -np 8'.
            Note that PolyChord must be installed with MPI enabled to allow
            running with MPI.
        """
        self.config_str = kwargs.pop('config_str', None)
        self.derived_str = kwargs.pop('derived_str', None)
        self.mpi_str = kwargs.pop('mpi_str', None)
        if kwargs:
            raise TypeError('unexpected **kwargs: {0}'.format(kwargs))
        self.executable_path = executable_path
        self.prior_str = prior_str

    def __call__(self, settings_dict, comm=None):
        """
        Run PolyChord with the input settings by writing a .ini file then using
        the compiled likelihood specified in executable_path.

        See the PolyChord documentation for more details.

        Parameters
        ----------
        settings_dict: dict
            Input PolyChord settings.
        comm: None, optional
            Not used. Included only so __call__ has the same arguments as the
            equivalent python function (which uses the comm argument for
            runnign with MPI).
        """
        assert os.path.isfile(self.executable_path), (
            'executable not found: ' + self.executable_path)
        assert comm is None, 'comm not used for compiled likelihoods.'
        # Write settings to ini file
        file_path = os.path.join(
            settings_dict['base_dir'], settings_dict['file_root'])
        with open(file_path + '.ini', 'w') as ini_file:
            ini_file.write(self.ini_string(settings_dict))
        # If required, write config file
        if self.config_str is not None:
            with open(file_path + '.cfg', 'w') as cfg_file:
                cfg_file.write(self.config_str)
        # Execute command
        command_str = self.executable_path + ' ' + file_path + '.ini'
        if self.mpi_str is not None:
            command_str = self.mpi_str + ' ' + command_str
        os.system(command_str)

    def ini_string(self, settings):
        """Get a PolyChord format .ini file string based on the input settings.

        Parameters
        ----------
        settings: dict

        Returns
        -------
        string: str
        """
        string = ''
        # Add the settings
        for key, value in settings.items():
            if key == 'nlives':
                if value:
                    loglikes = sorted(settings['nlives'])
                    string += 'loglikes = ' + format_setting(loglikes) + '\n'
                    nlives = [settings['nlives'][ll] for ll in loglikes]
                    string += 'nlives = ' + format_setting(nlives) + '\n'
            else:
                string += key + ' = ' + format_setting(value) + '\n'
        # Add the prior
        string += self.prior_str
        if self.derived_str is not None:
            string += self.derived_str
        return string


# Helper functions for making PolyChord prior strings
# ---------------------------------------------------


def get_prior_block_str(prior_name, prior_params, nparam, **kwargs):
    """
    Returns a PolyChord format prior block for inclusion in PolyChord .ini
    files.

    See the PolyChord documentation for more details.

    Parameters
    ----------
    prior_name: str
        Name of prior. See the PolyChord documentation for a list of currently
        available priors and details of how to add your own.
    prior_params: str, float or list of strs and floats
        Parameters for the prior function.
    nparam: int
        Number of parameters.
    start_param: int, optional
        Where to start param numbering. For when multiple prior blocks are
        being used.
    block: int, optional
        Number of block (only needed when using multiple prior blocks).
    speed: int, optional
        Use to specify fast and slow parameters if required. See the PolyChord
        documentation for more details.

    Returns
    -------
    block_str: str
        PolyChord format prior block string for ini file.
    """
    start_param = kwargs.pop('start_param', 1)
    speed = kwargs.pop('speed', 1)
    block = kwargs.pop('block', 1)
    if kwargs:
        raise TypeError('unexpected **kwargs: {0}'.format(kwargs))
    block_str = ''
    for i in range(start_param, nparam + start_param):
        block_str += ('P : p{0} | \\theta_{{{0}}} | {1} | {2} | {3} |'
                      .format(i, speed, prior_name, block))
        block_str += format_setting(prior_params) + '\n'
    return block_str


def format_setting(setting):
    """
    Return setting as string in the format needed for PolyChord's .ini files.
    These use 'T' for True and 'F' for False, and require lists of numbers
    written separated by spaces and without commas or brackets.

    Parameters
    ----------
    setting: (can be any type for which str(settings) works)

    Returns
    -------
    str
    """
    if isinstance(setting, bool):
        return str(setting)[0]
    elif isinstance(setting, (list, tuple)):
        string = str(setting)
        for char in [',', '[', ']', '(', ')']:
            string = string.replace(char, '')
        return string
    else:
        return str(setting)


def python_prior_to_str(prior, **kwargs):
    """Utility function for mapping python priors (of the type in
    python_priors.py) to  ini file format strings used for compiled (C++
    or Fortran) likelihoods.

    The input prior must correspond to a prior function set up in
    PolyChord/src/polychord/priors.f90. You can easily add your own too.
    Note that some of the priors are only available in PolyChord >= v1.15.

    Parameters
    ----------
    prior_obj: python prior object
        Of the type defined in python_priors.py
    kwargs: dict, optional
        Passed to get_prior_block_str (see its docstring for more details).

    Returns
    -------
    block_str: str
        PolyChord format prior block string for ini file.
    """
    nparam = kwargs.pop('nparam')
    name = type(prior).__name__.lower()
    if name == 'uniform':
        parameters = [prior.minimum, prior.maximum]
    elif name == 'poweruniform':
        name = 'power_uniform'
        parameters = [prior.minimum, prior.maximum, prior.power]
        assert prior.power < 0, (
            'compiled power_uniform currently only takes negative powers.'
            'power={}'.format(prior.power))
    elif name == 'gaussian':
        parameters = [getattr(prior, 'mu', 0.0), prior.sigma]
        if getattr(prior, 'half', False):
            name = 'half_' + name
    elif name == 'exponential':
        parameters = [prior.lambd]
    else:
        raise TypeError('Not set up for ' + name)
    if getattr(prior, 'sort', False):
        name = 'sorted_' + name
    if getattr(prior, 'adaptive', False):
        name = 'adaptive_' + name
        assert getattr(prior, "nfunc_min", 1) == 1, (
            'compiled adaptive priors currently only take nfunc_min=1.'
            'prior.nfunc_min={}'.format(prior.nfunc_min))
    return get_prior_block_str(name, parameters, nparam, **kwargs)


def python_block_prior_to_str(bp_obj):
    """As for python_prior_to_str, but for BlockPrior objects of the type
    defined in python_priors.py. python_prior_to_str is called seperately on
    every block.

    Parameters
    ----------
    prior_obj: python prior object
        Of the type defined in python_priors.py.
    kwargs: dict, optional
        Passed to get_prior_block_str (see its docstring for more details).

    Returns
    -------
    block_str: str
        PolyChord format prior block string for ini file.
    """
    assert type(bp_obj).__name__ == 'BlockPrior', (
        'Unexpected input object type: {}'.format(
            type(bp_obj).__name__))
    start_param = 1
    string = ''
    for i, prior in enumerate(bp_obj.prior_blocks):
        string += python_prior_to_str(
            prior, block=(i + 1), start_param=start_param,
            nparam=bp_obj.block_sizes[i])
        start_param += bp_obj.block_sizes[i]
    return string



#!/usr/bin/env python
"""
Functions for loading and processing dyPolyChord dynamic nested sampling runs.
"""
import os
import warnings
import numpy as np
import nestcheck.ns_run_utils
import nestcheck.data_processing
import nestcheck.io_utils as iou


def settings_root(likelihood_name, prior_name, ndim, **kwargs):
    """
    Returns a standard string containing information about settings.

    Parameters
    ----------
    likelihood_name: str
    prior_name: str
    ndim: int
    prior_scale: float or int
    dynamic_goal: float, int or None
    nlive_const: int
    nrepeats: int
    nint: int, optional
    init_step: int, optional

    Returns
    -------
    root: str
    """
    prior_scale = kwargs.pop('prior_scale')
    dynamic_goal = kwargs.pop('dynamic_goal')
    nlive_const = kwargs.pop('nlive_const')
    nrepeats = kwargs.pop('nrepeats')
    ninit = kwargs.pop('ninit', None)
    init_step = kwargs.pop('init_step', None)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    root = '{}_{}_{}_dg{}'.format(likelihood_name, prior_name, prior_scale,
                                  dynamic_goal)
    if dynamic_goal is not None:
        assert ninit is not None
        root += '_{}init'.format(ninit)
        if dynamic_goal != 0:
            assert init_step is not None
            root += '_{}is'.format(init_step)
    root += '_{}d_{}nlive_{}nrepeats'.format(ndim, nlive_const, nrepeats)
    return root.replace('.', '_')


def process_dypolychord_run(file_root, base_dir, **kwargs):
    """
    Load the output files of a dynamic run and process them to the nestcheck
    format.

    Parameters
    ----------
    file_root: str
    base_dir: str
    dynamic_goal: float
    dup_assert: bool, optional
        Whether to throw an AssertionError if there are duplicate point
        loglikelihood values.
    dup_warn: bool, optional
        Whether to give a UserWarning if there are duplicate point
        loglikelihood values.

    Returns
    -------
    run: dict
        Nested sampling run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    """
    dynamic_goal = kwargs.pop('dynamic_goal')
    dup_assert = kwargs.pop('dup_assert', False)
    dup_warn = kwargs.pop('dup_warn', False)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    init = nestcheck.data_processing.process_polychord_run(
        file_root + '_init', base_dir, dup_assert=dup_assert,
        dup_warn=dup_warn)
    assert np.all(init['thread_min_max'][:, 0] == -np.inf), (
        'Initial run contains threads not starting at -inf.\n'
        'thread_min_max=' + str(init['thread_min_max']))
    dyn = nestcheck.data_processing.process_polychord_run(
        file_root + '_dyn', base_dir, dup_assert=dup_assert,
        dup_warn=dup_warn)
    dyn_info = iou.pickle_load(os.path.join(
        base_dir, file_root + '_dyn_info'))
    if dynamic_goal == 0:
        # If dynamic_goal == 0 then nlive should only decrease, so check all
        # threads start by sampling
        assert np.all(dyn['thread_min_max'][:, 0] == -np.inf), (
            str(dyn['thread_min_max']))
    # Get info to run
    run_output = {'file_root': file_root,
                  'base_dir': base_dir}
    if 'resume_ndead' not in dyn_info:
        # The dynamic run was not resumed part way through the initial run:
        # hence there are no samples repeated in both runs' files and we can
        # simply combine dyn and init using standard nestcheck functions.
        run = nestcheck.ns_run_utils.combine_ns_runs([init, dyn])
        try:
            run_output['nlike'] = (
                init['output']['nlike'] + dyn['output']['nlike'])
        except KeyError:
            pass # protect from error reading nlike from .stats file
    else:
        # The dynamic run started by resuming the initial run part way
        # through, and we need to remove duplicate points from the combined run
        run = combine_resumed_dyn_run(init, dyn, dyn_info['resume_ndead'])
        try:
            run_output['nlike'] = (
                init['output']['nlike'] + dyn['output']['nlike']
                - dyn_info['resume_nlike'])
        except KeyError:
            pass # protect from error reading nlike from .stats file
    run['output'] = run_output
    # check the nested sampling run has the expected properties
    nestcheck.ns_run_utils.check_ns_run(
        run, dup_assert=dup_assert, dup_warn=dup_warn)
    return run


def combine_resumed_dyn_run(init, dyn, resume_ndead):
    """
    Merge initial run and dynamic run which was resumed from it, including
    removing duplicate points present in both runs.

    Parameters
    ----------
    init: dict
        Initial exploratory run in nestcheck format (see
        http://nestcheck.readthedocs.io/en/latest/api.html for more
        information).
    dyn: dict
        Dynamic run in nestcheck format.
    resume_ndead: int
        The number of dead points present when dyn was resumed from init.

    Returns
    -------
    run: dict
        Combined run in nestcheck format.
    """
    assert np.array_equal(
        init['logl'][:resume_ndead], dyn['logl'][:resume_ndead]), (
            'The first {0} points should be the same'.format(resume_ndead))
    init['theta'] = init['theta'][resume_ndead:, :]
    for key in ['nlive_array', 'logl', 'thread_labels']:
        init[key] = init[key][resume_ndead:]
    # We also need to remove the points that were live when the resume file was
    # written, as these show up as samples in both dyn and init
    live_inds = []
    empty_thread_inds = []
    for i, th_lab in enumerate(np.unique(init['thread_labels'])):
        th_inds = np.where(init['thread_labels'] == th_lab)[0]
        live_logl = init['logl'][th_inds[0]]
        init['thread_min_max'][i, 0] = live_logl
        if np.where(dyn['logl'] == live_logl)[0].shape[0] > 0:
            live_inds.append(th_inds[0])
            if th_inds.shape[0] == 1:
                empty_thread_inds.append(i)
        else:
            warnings.warn(
                ('Expected live point at resume should be present in dynamic '
                 'run. If there are no further errors, this warning can be '
                 'ignored.\nlogl={}, th_lab={}, inds={}, init samples (after '
                 'removing first resume_ndead)={}, unique threads in init={}, '
                 'dyn samples={}, resume_ndead={}.').format(
                     str(live_logl), th_lab,
                     np.where(dyn['logl'] == live_logl),
                     init['logl'].shape[0],
                     np.unique(init['thread_labels']).shape[0],
                     dyn['logl'].shape[0], resume_ndead), UserWarning)
    # Remove the live points at resume from init
    init['theta'] = np.delete(init['theta'], live_inds, axis=0)
    for key in ['nlive_array', 'logl', 'thread_labels']:
        init[key] = np.delete(init[key], live_inds)
    # Deal with the case that one of the threads is now empty
    if empty_thread_inds:
        # remove any empty threads from logl_min_max
        init['thread_min_max'] = np.delete(
            init['thread_min_max'], empty_thread_inds, axis=0)
        # Now we need to reorder the thread labels to avoid gaps
        thread_labels_new = np.full(init['thread_labels'].shape, np.nan)
        for i, th_lab in enumerate(np.unique(init['thread_labels'])):
            inds = np.where(init['thread_labels'] == th_lab)[0]
            thread_labels_new[inds] = i
            # Check the newly relabelled thread label matches thread_min_max
            assert init['thread_min_max'][i, 0] <= init['logl'][inds[0]]
            assert init['thread_min_max'][i, 1] == init['logl'][inds[-1]]
        assert np.all(~np.isnan(thread_labels_new))
        init['thread_labels'] = thread_labels_new.astype(int)
    # Add the init threads to dyn with new labels that continue on from the dyn
    # labels
    init['thread_labels'] += dyn['thread_min_max'].shape[0]
    run = nestcheck.ns_run_utils.combine_threads(
        nestcheck.ns_run_utils.get_run_threads(dyn) +
        nestcheck.ns_run_utils.get_run_threads(init),
        assert_birth_point=False)
    return run



#!/usr/bin/env python
"""
Loglikelihood functions for use with PolyChord's python interface.

PolyChord >= v1.14 requires likelihoods to be callables with parameter and
return signatures:

Parameters
----------
theta: float or 1d numpy array

Returns
-------
logl: float
    Loglikelihood.
phi: list of length nderived
    Any derived parameters.

We use classes with the loglikelihood defined in the __call__ property, as
this provides convenient way of storing other information such as
hyperparameter values. These objects can be used in the same way as functions
due to python's "duck typing" (alternatively you can define likelihoods
using functions).
"""
import copy
import numpy as np
import scipy.special


class Gaussian(object):

    """Symmetric Gaussian loglikelihood centered on the origin."""

    def __init__(self, sigma=1.0, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        sigma: float, optional
            Standard deviation of Gaussian (the same for each parameter).
        nderived: int, optional
            Number of derived parameters.
        """
        self.sigma = sigma
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        logl = log_gaussian_pdf(theta, sigma=self.sigma, mu=0)
        return logl, [0.0] * self.nderived


class GaussianShell(object):

    """Gaussian Shell loglikelihood centred on the origin."""

    def __init__(self, sigma=0.2, rshell=2, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        sigma: float, optional
            Standard deviation of Gaussian (the same for each parameter).
        rshell: float, optional
            Distance of shell peak from origin.
        nderived: int, optional
            Number of derived parameters.
        """
        self.sigma = sigma
        self.rshell = rshell
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        rad = np.sqrt(np.sum(theta ** 2))
        logl = - ((rad - self.rshell) ** 2) / (2 * self.sigma ** 2)
        return logl, [0.0] * self.nderived


class Rastrigin(object):

    """Rastrigin loglikelihood as described in "PolyChord: next-generation
    nested sampling" (Handley et al., 2015)."""

    def __init__(self, a=10, nderived=0):
        """
        Set up likelihood object's hyperparameter values.

        Parameters
        ----------
        a: float, optional
        nderived: int, optional
            Number of derived parameters.
        """
        self.a = a
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        ftheta = self.a * len(theta)
        for th in theta:
            ftheta += (th ** 2) - self.a * np.cos(2 * np.pi * th)
        logl = -ftheta
        return logl, [0.0] * self.nderived


class Rosenbrock(object):

    """Rosenbrock loglikelihood as described in "PolyChord: next-generation
    nested sampling" (Handley et al., 2015)."""

    def __init__(self, a=1, b=100, nderived=0):
        """
        Define likelihood hyperparameter values.

        Parameters
        ----------
        theta: 1d numpy array
            Parameters.
        a: float, optional
        b: float, optional
        nderived: int, optional
            Number of derived parameters.
        """
        self.a = a
        self.b = b
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        ftheta = 0
        for i in range(len(theta) - 1):
            ftheta += (self.a - theta[i]) ** 2
            ftheta += self.b * ((theta[i + 1] - (theta[i] ** 2)) ** 2)
        logl = -ftheta
        return logl, [0.0] * self.nderived


class GaussianMix(object):

    r"""Gaussian mixture likelihood in :math:`\ge 2` dimensions with up to
    4 compoents.

    Each component has the same standard deviation :math:`\sigma`, and
    they their centres respectively have :math:`(\theta_1, \theta_2)`
    coordinates:

    (0, sep), (0, -sep), (sep, 0), (-sep, 0).
    """

    def __init__(self, sep=4, weights=(0.4, 0.3, 0.2, 0.1), sigma=1,
                 nderived=0):
        """
        Define likelihood hyperparameter values.

        Parameters
        ----------
        sep: float
            Distance from each Gaussian to the origin.
        weights: iterable of floats
            weights of each Gaussian component.
        sigma: float
            Stanard deviation of Gaussian components.
        nderived: int, optional
            Number of derived parameters.
        """
        assert len(weights) in [2, 3, 4], (
            'Weights must have 2, 3 or 4 components. Weights=' + str(weights))
        assert np.isclose(sum(weights), 1), (
            'Weights must sum to 1! Weights=' + str(weights))
        self.nderived = nderived
        self.weights = weights
        self.sigmas = [sigma] * len(weights)
        positions = []
        positions.append(np.asarray([0, sep]))
        positions.append(np.asarray([0, -sep]))
        positions.append(np.asarray([sep, 0]))
        positions.append(np.asarray([-sep, 0]))
        self.positions = positions[:len(weights)]

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        thetas = []
        for pos in self.positions:
            thetas.append(copy.deepcopy(theta))
            thetas[-1][:2] -= pos
        logls = [(Gaussian(sigma=self.sigmas[i])(thetas[i])[0]
                  + np.log(self.weights[i])) for i in range(len(self.weights))]
        logl = scipy.special.logsumexp(logls)
        return logl, [0.0] * self.nderived


class LogGammaMix(object):

    """The loggamma mix used in Beaujean and Caldwell (2013) and Feroz et al.
    (2013)."""

    def __init__(self, nderived=0):
        """
        Define likelihood hyperparameter values.

        Parameters
        ----------
        nderived: int, optional
            Number of derived parameters.
        """
        self.nderived = nderived

    def __call__(self, theta):
        """
        Calculate loglikelihood(theta), as well as any derived parameters.

        N.B. this loglikelihood is not normalised.

        Parameters
        ----------
        theta: float or 1d numpy array

        Returns
        -------
        logl: float
            Loglikelihood
        phi: list of length nderived
            Any derived parameters.
        """
        assert theta.shape[0] % 2 == 0, (
            'ndim={} must be even'.format(theta.shape))
        # first component is
        # l(t) = 0.5 * loggamma(t - 10, 1, 1) + 0.5 * loggamma(t + 10, 1, 1)
        logl = np.log(0.5) + scipy.special.logsumexp(
            [log_loggamma_pdf(th, alpha=1, beta=1) for th in
             [theta[0] - 10, theta[0] + 10]])
        # second component is
        # l(t) = 0.5 * N(t - 10, 1, 1) + 0.5 * N(t + 10, 1, 1)
        logl += np.log(0.5) + scipy.special.logsumexp(
            [log_gaussian_pdf(th, sigma=1) for th in
             [theta[1] - 10, theta[1] + 10]])
        if theta.shape[0] > 2:
            # remaining components are half loggamma and half gaussian
            bound_ind = (theta.shape[0] // 2) + 1
            logl += log_loggamma_pdf(theta[2:bound_ind], alpha=1, beta=1)
            logl += log_gaussian_pdf(theta[bound_ind:], sigma=1)
        return logl, [0.0] * self.nderived


# Helper functions
# ----------------

def log_loggamma_pdf_1d(theta, alpha=1, beta=1):
    r"""1d gamma distribution, with each component of theta independently
    having PDF:

    .. math::

        f(x) = \frac{
            \mathrm{e}^{\beta x} \mathrm{e}^{-\mathrm{e}^x / \alpha}
            }{\alpha^{\beta} \Gamma(\beta)}

    This function returns :math:`\log f(x)`.

    Parameters
    ----------
    theta: float or array
        Where to evaluate :math:`\log f(x)`. Values
        :math:`\in (-\inf, \inf)`.
    alpha: float, > 0
        Scale parameter
    beta: float, > 0
        Shape parameter

    Returns
    -------
    logl: same type as theta
    """
    # log of numerator
    logl = beta * theta
    logl += -np.exp(theta) / alpha
    # log of denominator
    logl -= np.log(alpha) * beta
    logl -= scipy.special.gammaln(beta)
    return logl


def log_loggamma_pdf(theta, alpha=1, beta=1):
    """Loglikelihood for multidimensional loggamma distribution, with
    each component of theta having an independent loggamma distribution.

    Always returns a float.

    Parameters
    ----------
    theta: float or numpy array
    alpha: float, optional
    beta: float, optional

    Returns
    -------
    logl: float
        Loglikelihood.
    """
    logl = log_loggamma_pdf_1d(theta, alpha=alpha, beta=beta)
    # If there are many components, sum their log-space consributions
    if not isinstance(logl, (float, int)):
        assert logl.shape == theta.shape, (
            'logl={} theta={}'.format(logl, theta))
        logl = np.sum(logl)
    return logl


def log_gaussian_pdf(theta, sigma=1, mu=0, ndim=None):
    """Log of uncorrelated Gaussian pdf.

    Parameters
    ----------
    theta: float or 1d numpy array
    sigma: float, optional
    mu: float, optional
    ndim: int, optional

    Returns
    -------
    logl: float
        Loglikelihood.
    """
    if ndim is None:
        try:
            ndim = len(theta)
        except TypeError:
            assert isinstance(theta, (float, int)), theta
            ndim = 1
    logl = -(np.sum((theta - mu) ** 2) / (2 * sigma ** 2))
    logl -= np.log(2 * np.pi * (sigma ** 2)) * ndim / 2.0
    return logl



#!/usr/bin/env python
"""Python priors for use with PolyChord.

PolyChord v1.14 requires priors to be callables with parameter and return
types:

Parameters
----------
hypercube: float or 1d numpy array
    Parameter positions in the prior hypercube.

Returns
-------
theta: float or 1d numpy array
    Corresponding physical parameter coordinates.

Input hypercube values numpy array is mapped to physical space using the
inverse CDF (cumulative distribution function) of each parameter.
See the PolyChord papers for more details.

This module use classes with the prior defined in the cube_to_physical
function and called with __call__ property. This provides convenient way of
storing other information such as hyperparameter values. The objects be
used in the same way as functions due to python's "duck typing" (or
alternatively you can just define prior functions).

The BlockPrior class allows convenient use of different priors on different
parameters.

Inheritance of the BasePrior class allows priors to:

   1. have parameters' values sorted to give an enforced order. Useful when the
   parameter space is symmetric under exchange of variables as this allows the
   space to be explored to be contracted by a factor of N! (where N is the
   number of such parameters);

   2. adaptively select the number of parameters to use.

You can ignore these if you don't need them.
"""
import copy
import numpy as np
import scipy


class BasePrior(object):

    """Base class for Priors."""

    def __init__(self, adaptive=False, sort=False, nfunc_min=1):
        """
        Set up prior object's hyperparameter values.

        Parameters
        ----------
        adaptive: bool, optional
        sort: bool, optional
        nfunc_min: int, optional
        """
        self.adaptive = adaptive
        self.sort = sort
        self.nfunc_min = nfunc_min

    def cube_to_physical(self, cube):  # pylint: disable=no-self-use
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        cube: 1d numpy array
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.

        Returns
        -------
        theta: 1d numpy array
            Physical parameter values corresponding to hypercube.
        """
        return cube

    def __call__(self, cube):
        """
        Evaluate prior on hypercube coordinates.

        Parameters
        ----------
        cube: 1d numpy array
            Point coordinate on unit hypercube (in probabily space).
            Note this variable cannot be edited else PolyChord throws an error.

        Returns
        -------
        theta: 1d numpy array
            Physical parameter values for prior.
        """
        if self.adaptive:
            try:
                theta = adaptive_transform(
                    cube, sort=self.sort, nfunc_min=self.nfunc_min)
            except ValueError:
                if np.isnan(cube[0]):
                    return np.full(cube.shape, np.nan)
                else:
                    raise
            theta[1:] = self.cube_to_physical(theta[1:])
            return theta
        else:
            if self.sort:
                return self.cube_to_physical(forced_identifiability(cube))
            else:
                return self.cube_to_physical(cube)


class Gaussian(BasePrior):

    """Symmetric Gaussian prior centred on the origin."""

    def __init__(self, sigma=10.0, half=False, mu=0.0, **kwargs):
        """
        Set up prior object's hyperparameter values.

        Parameters
        ----------
        sigma: float
            Standard deviation of Gaussian prior in each parameter.
        half: bool, optional
            Half-Gaussian prior - nonzero only in the region greater than mu.
            Note that in this case mu is no longer the mean and sigma is no
            longer the standard deviation of the prior.
        mu: float, optional
            Mean of Gaussian prior.
        kwargs: dict, optional
            See BasePrior.__init__ for more infomation.
        """
        BasePrior.__init__(self, **kwargs)
        self.sigma = sigma
        self.mu = mu
        self.half = half

    def cube_to_physical(self, cube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        cube: 1d numpy array
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.

        Returns
        -------
        theta: 1d numpy array
            Physical parameter values corresponding to hypercube.
        """
        if self.half:
            theta = scipy.special.erfinv(cube)
        else:
            theta = scipy.special.erfinv(2 * cube - 1)
        theta *= self.sigma * np.sqrt(2)
        return theta + self.mu


class Uniform(BasePrior):

    """Uniform prior."""

    def __init__(self, minimum=0.0, maximum=1.0, **kwargs):
        """
        Set up prior object's hyperparameter values.

        Prior is uniform in [minimum, maximum] in each parameter.

        Parameters
        ----------
        minimum: float
        maximum: float
        kwargs: dict, optional
            See BasePrior.__init__ for more infomation.
        """
        BasePrior.__init__(self, **kwargs)
        assert maximum > minimum, (minimum, maximum)
        self.maximum = maximum
        self.minimum = minimum

    def cube_to_physical(self, cube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        cube: 1d numpy array

        Returns
        -------
        theta: 1d numpy array
        """
        return self.minimum + (self.maximum - self.minimum) * cube


class PowerUniform(BasePrior):

    """Uniform in theta ** power"""

    def __init__(self, minimum=0.1, maximum=2.0, power=-2, **kwargs):
        """
        Set up prior object's hyperparameter values.

        Prior is uniform in [minimum, maximum] in each parameter.

        Parameters
        ----------
        minimum: float
        maximum: float
        power: float or int
        kwargs: dict, optional
            See BasePrior.__init__ for more infomation.
        """
        BasePrior.__init__(self, **kwargs)
        assert maximum > minimum > 0, (minimum, maximum)
        assert power != 0, power
        self.maximum = maximum
        self.minimum = minimum
        self.power = power
        self.const = abs((minimum ** (1. / power)) - (maximum ** (1. / power)))
        self.const = 1 / self.const

    def cube_to_physical(self, cube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        cube: 1d numpy array

        Returns
        -------
        theta: 1d numpy array
        """
        if self.power > 0:
            theta = (self.minimum ** (1. / self.power)) + (cube / self.const)
        else:
            theta = (self.minimum ** (1. / self.power)) - (cube / self.const)
        return theta ** self.power


class Exponential(BasePrior):

    """Exponential prior."""

    def __init__(self, lambd=1.0, **kwargs):
        """
        Set up prior object's hyperparameter values.

        Parameters
        ----------
        lambd: float
        kwargs: dict, optional
            See BasePrior.__init__ for more infomation.
        """
        BasePrior.__init__(self, **kwargs)
        self.lambd = lambd

    def cube_to_physical(self, cube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        cube: 1d numpy array

        Returns
        -------
        theta: 1d numpy array
        """
        return - np.log(1 - cube) / self.lambd


class BlockPrior(object):

    """Prior object which applies a list of priors to different blocks within
    the parameters."""

    def __init__(self, prior_blocks, block_sizes):
        """Store prior and size of each block."""
        assert len(prior_blocks) == len(block_sizes), (
            'len(prior_blocks)={}, len(block_sizes)={}, block_sizes={}'
            .format(len(prior_blocks), len(block_sizes), block_sizes))
        self.prior_blocks = prior_blocks
        self.block_sizes = block_sizes

    def __call__(self, cube):
        """
        Map hypercube values to physical parameter values.

        Parameters
        ----------
        hypercube: 1d numpy array
            Point coordinate on unit hypercube (in probabily space).
            See the PolyChord papers for more details.

        Returns
        -------
        theta: 1d numpy array
            Physical parameter values corresponding to hypercube.
        """
        theta = np.zeros(cube.shape)
        start = 0
        end = 0
        for i, prior in enumerate(self.prior_blocks):
            end += self.block_sizes[i]
            theta[start:end] = prior(cube[start:end])
            start += self.block_sizes[i]
        return theta


# Helper functions
# ----------------


def forced_identifiability(cube):
    """Transform hypercube coordinates to enforce identifiability.

    For more details see: "PolyChord: next-generation nested sampling"
    (Handley et al. 2015).

    Parameters
    ----------
    cube: 1d numpy array
        Point coordinate on unit hypercube (in probabily space).

    Returns
    -------
    ordered_cube: 1d numpy array
    """
    ordered_cube = np.zeros(cube.shape)
    ordered_cube[-1] = cube[-1] ** (1. / cube.shape[0])
    for n in range(cube.shape[0] - 2, -1, -1):
        ordered_cube[n] = cube[n] ** (1. / (n + 1)) * ordered_cube[n + 1]
    return ordered_cube


def adaptive_transform(cube, sort=True, nfunc_min=1):
    """Tranform first parameter (nfunc) to uniform in (nfunc_min, nfunc_max)
    and, if required, perform forced identifiability transform on the next
    nfunc parameters only.

    Parameters
    ----------
    cube: 1d numpy array
        Point coordinate on unit hypercube (in probabily space).

    Returns
    -------
    ad_cube: 1d numpy array
        First element is physical coordinate of nfunc parameter, other elements
        are cube coordinates with any forced identifiability transform already
        applied.
    """
    # First get integer number of funcs
    ad_cube = copy.deepcopy(cube)
    nfunc_max = cube.shape[0] - 1
    # first component is a number of funcs
    ad_cube[0] = ((nfunc_min - 0.5) + (1.0 + nfunc_max - nfunc_min) * cube[0])
    if sort:
        nfunc = int(np.round(ad_cube[0]))
        # Sort only parameters 1 to nfunc
        ad_cube[1:1 + nfunc] = forced_identifiability(cube[1:1 + nfunc])
    return ad_cube
